/*                  - Mellanox Confidential and Proprietary -
 *
 *  Copyright (C) 2010-2011, Mellanox Technologies Ltd.  ALL RIGHTS RESERVED.
 *
 *  Except as specifically permitted herein, no portion of the information,
 *  including but not limited to object code and source code, may be reproduced,
 *  modified, distributed, republished or otherwise exploited in any form or by
 *  any means for any purpose without the prior written permission of Mellanox
 *  Technologies Ltd. Use of software subject to the terms and conditions
 *  detailed in the file "COPYING".
 *
 */

#ifndef CC_MGR_H_
#define CC_MGR_H_
/* osm_config.h needs to be before all other includes */
#include <opensm/osm_config.h>
#include <sys/time.h>
#include <math.h>
#include <unistd.h>
#include <typeinfo>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <iterator>
#include <algorithm>

#include <opensm/osm_opensm.h>
#include <opensm/osm_helper.h>
#include <opensm/osm_sa.h>
#include <opensm/osm_event_plugin.h>
#include <vendor/osm_vendor_api.h>

#include "fit_api.h"
#include "cc_mgr_types.h"


//============================================================================//
class CongestionControlManager {
public:
    ////////////////////
    // methods
    ////////////////////
    /**
     * Constructor CongestionControlManager
     *
     * @param p_osm pointer to OpenSM structure
     * @param osm_plugin_options osm plugin options
     */
    CongestionControlManager(osm_opensm_t *p_osm,
            char *osm_plugin_options);

    /**
     *  Destructor CongestionControlManager
     *
     */
    ~CongestionControlManager();

    /**
     * Set osm_event_id
     *
     * @param osm_event_id
     */
    void SetOSMEventId(osm_epi_event_id_t osm_event_id);

    /**
     * Update fabric nodes. Update node info for all nodes
     *
     * @return int 0 if succeeded, otherwise 1
     */
    int UpdateFabricNodes();

    /**
     * Configure all nodes the support congestion control in the
     * fabric with congestion control parameters with user defined
     * parameters or default if none where given.
     *
     * @return int return 1 is succeeded, otherwise 0
     */
    int Run();

    /**
     * Print Exception
     *
     * @param e exception
     * @param p_cc_osm_log log file
     */
    static void printException(exception& e, osm_log_t* p_cc_osm_log)
    {
    string e_what(e.what( ));
    string e_type(typeid(e).name( ));
    OSM_LOG(p_cc_osm_log, OSM_LOG_DEBUG,
            "CC_MGR - caught an exception: %s. Type: %s\n",
            e_what.c_str(), e_type.c_str());
    }

private:
    ////////////////////
    //members
    ////////////////////
    osm_epi_event_id_t          m_osm_event_id;         /** OSM event id */
    unsigned int                m_num_hosts;            /** Number of hosts */
    Ibis                        m_ibis_obj;             /** IBIS object, used to send mads */
    u_int64_t                   m_port_guid;            /** Port guid to use when binding port */
    osm_opensm_t                *m_p_osm;               /** Pointer to osm opensm struct */
    osm_vendor_t                *m_p_osm_vendor;        /** Pointer to osm vendor struct */
    osm_subn_t                  *m_p_osm_subn;          /** Pointer to osm subnet struct */
    osm_sa_t                    *m_p_osm_sa;            /** OSM subnet agent */
    osm_log_t                   *m_p_osm_log;           /** Pointer to osm log */
    u_int64_t                   m_cc_key;               /** CC Key to use when sending congestion control mads */
    string                      m_conf_file_name;       /** Name of user input file */
    unsigned int                m_max_errors;           /** Max errors allowed */
    long                        m_error_window;         /** Error window */
    int                         m_oldest_error;         /** index of oldest error in error window array */
    unsigned int                m_num_errors;           /** Number of errors */
    struct timeval              *m_p_error_window;      /** Error window array */
    bool                        m_enable;               /** Enable status for all nodes */
    bool                        m_cc_config_up;         /** congestion control configuration is up*/
    struct SWCCSettingDataBase  m_sw_cc_setting_db;     /** Switch congestion control settings database */
    struct CACCSettingDataBase  m_ca_cc_setting_db;     /** CA congestion control settings database */
    struct UserOptionDataBase   m_user_option_db;       /** User options database */
    struct MasterDataBase       m_master_db;            /** Master congestion control settings database */

    map<string, USER_OPTIONS_STR> m_ca_user_options; /** CA user options database */
    map<string, USER_OPTIONS_STR> m_sw_user_options; /** Switch user options database */
    map<string, USER_OPTIONS_STR> m_general_user_options; /** General user options database */

    volatile bool   m_cc_log_agent_created;
    volatile bool   m_run_cc_statistics_agent;
    volatile bool   m_cc_mgr_request;

    volatile uint8_t m_lock_holder;
    pthread_mutex_t  m_lock_holder_mutex;

    pthread_t       m_cc_log_agent_thread;
    pthread_mutex_t m_cc_mutex;
    pthread_mutex_t m_cc_gate_mutex;

    string          m_cc_log_file_name;
    long            m_cc_statistics_cycle;
    unsigned int    m_cc_log_file_size;
    map<u_int64_t, u_int16_t> m_ca_cc_log_event_cnt_db;
    map<u_int64_t, u_int16_t> m_sw_cc_log_event_cnt_db;

    CCStatisticsDataBase m_cc_statistics_db;

    ////////////////////
    // methods mutex
    ////////////////////
    void MgrLockCCMutex();
    void UnLockCCMutex(uint8_t lock_holder);
    void StatLockCCMutex();

    ////////////////////
    // methods CheckRC for MAD send
    ////////////////////
    /**
     * Check if more than m_max_errors occurred during the
     *        last m_error_window. If so, throw an execption
     * @throw Throws int when condition holds
     * @param rc return value that indicates if an error has occurred
     */
    void CheckRC(int& rc);

    ////////////////////
    // methods parse osm params
    ////////////////////
    /**
     * Parse OSM parameters for cc_mgr conf file
     *
     * @param osm_params
     *
     * @return int 0 if succeeded and flag found, otherwise 1
     */
    int ParseConfFileName(char *osm_plugin_options);

    ////////////////////
    // methods construct
    ////////////////////
    /**
     * Initialize user options. Add entry for each user
     *        option string key
     *
     */
    void InitUserOptions();

    /**
     * Set defualt values to master database
     *
     */
    void InitMasterDataBase();

    /**
     * Additional actions on instantiation
     *
     * @param osm_params
     *
     */
    void Construct(char *osm_plugin_options);

    ////////////////////
    // methods remove old db
    ////////////////////
    /**
     * Check if node exists in subnet (m_p_osm_subn) and
     *        return it's type
     *
     * @param node_guid node guid of the node to check
     * @param node_type node type of the node that was checked
     *
     * @return bool Return true if node was found in subnet,
     *         otherwise false
     */
    bool CheckNodeType(const u_int64_t node_guid, u_int8_t& node_type);

    /**
     * Remove nodes that don't exist in m_p_osm_subn from CA
     *        database and Switch database
     */
    void RemoveAbsentNodes();

    ////////////////////
    // methods init
    ////////////////////
    /**
     * Initialize objects.
     * Bind port to congestion control mads
     * Remove none existing nodes from database
     *
     * @return int 0 if succeeded, otherwise throw exception with value 1
     */
    int Init();

    ////////////////////
    // methods update db
    ////////////////////
    /**
     * Check if device id is supported.
     *
     * @param node_info node info holds device id
     *
     * @return bool true if supported, otherwise false
     */
    bool IsDeviceIDSupported(const CCNodeInfo& node_info);

    /**
     * Get from osm if the node should be updated.
     *
     * @param node_info node info
     * @param osm_update_needed
     *
     * @return int
     */
    int GetNodeOSMUpdateNeeded(const CCNodeInfo& node_info,
            bool& osm_update_needed);

    /**
     * Get nodes mtu size in bytes
     *
     * @param node_info node info
     * @param mtu_size mtu size, mtu size will be set here
     *
     * @return int 0 if succeeded, otherwise 1
     */
    int GetNodeMTUSize(const CCNodeInfo& node_info, u_int32_t& mtu_size);

    /**
     * Update CA node from sm database to local database ( lid, sl,
     * mtu_size)
     *
     * @param node_info node info
     *
     * @return int 0 if succeeded, otherwise 1
     */
    int UpdateCANode(const CCNodeInfo& node_info);

    /**
     * Update Switch node from sm database to local database ( lid,
     * sl)
     *
     * @param node_info node info
     *
     * @return int 0 if succeeded, otherwise 1
     */
    int UpdateSWNode(const CCNodeInfo& node_info);

    /**
     * Get correct sl
     *
     * @param p_dest_port destination port
     *
     * @return uint8_t sl
     */
    uint8_t GetSL(osm_port_t *p_dest_port);

    ////////////////////
    // methods cc support
    ////////////////////
    /**
     * Send MAD to check if node supports congestion control
     *
     * @param node_info node info
     * @param node_supports_cc true if node supports congestion
     *                         control, otherwise false
     *
     * @return int 0 if succeeded, otherwise 1
     */
    int CheckNodeSupportsCCoFabric(const CCNodeInfo& node_info,
            bool& node_supports_cc);

    /**
     * Check if CA node supports congestion control.
     *
     * If current state for this node in database is
     * CC_SUPPORT_UNKNOWN will send MAD to retrieve information
     *
     * @param node_info node info
     * @param node_supports_cc true if node supports congestion
     *                         control, otherwise false
     *
     * @return int 0 if succeeded, otherwise 1
     */
    int GetCANodeCCSupport(const CCNodeInfo& node_info,
            bool& node_supports_cc);

    /**
     * Check if Switch node supports congestion control.
     *
     * If current state for this node in database is
     * CC_SUPPORT_UNKNOWN will send MAD to retrieve information
     *
     * @param node_info
     * @param node_supports_cc
     *
     * @return int
     */
    int GetSWNodeCCSupport(const CCNodeInfo& node_info,
            bool& node_supports_cc);

    /**
     * Check all nodes if they support congestion control.
     * Indication will be set in database
     *
     *
     * @return int 0 if succeeded, otherwise 1
     */
    int CollectCCSupportForAllNodes();

    ////////////////////
    // methods cc configure
    ////////////////////
    /**
     * Get CA control_table_cap. Updated value in local database
     * Sends ControlTable MAD
     *
     * @param node_info node info
     * @param control_table_cap control table cap
     *
     * @return int 0 if succeeded, otherwise 1
     */
    int GetCAControlTableCap(const CCNodeInfo& node_info,
            unsigned int& control_table_cap);
    /**
     * Get CA control_table_cap for all nodes that support
     * congestion control.
     *
     */
    void GetCAControlTableCapForAllNodes();

    /**
     * Prints to osm log the content of the given CC_CACongestionSetting struct
     *
     * @param cc_ca_congestion_setting struct to be printed
     */
    void DumpCACongSetting(struct CC_CACongestionSetting& cc_ca_congestion_setting);

    /**
     * Calculate CCTI timer
     *
     * @param node_info node info
     * @param mtu_size node mtu size
     * @param ccti_timer calculated CCTI timer
     *
     * @return int 0 if succeeded, otherwise 1
     */
    int CalculateCCTITimer(const CCNodeInfo& node_info,
            const u_int32_t mtu_size,
            u_int16_t& ccti_timer);

    /**
      * Get settings for a specific CA node. Content will be
      *        filled in the given structs
      *
      * @param node_info node info
      * @param node_supports_cc true if node supports congestion
      *                         control, otherwise false
      * @param node_params CC_CACongestionSetting struct with node
      *                    settings (only if succeeded)
      *
      * @return int 0 if succeeded, otherwise 1
      */
     int GetCANodeCCSettings(const CCNodeInfo& node_info,
             bool& node_supports_cc,
             struct CC_CACongestionSetting& node_params,
             map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator& ca_node_it);

     /**
      * Compare CC_CACongestionSetting
      *
      * @param lhs
      * @param rhs
      *
      * @return bool true if all fields are equal, otherwise false
      */
     bool CompareCACongestionSetting(const struct CC_CACongestionSetting &lhs,
             const struct CC_CACongestionSetting &rhs);

     /**
      * Set CA CongestionSetting
      * Sends CACongestionSetting MAD
      *
      * @param node_info node info
      *
      * @return int 0 if succeeded, otherwise 1
      */
     int SetCACongestionSetting(const CCNodeInfo& node_info);

     /**
      * Set CA CongestionSetting for all nodes that support
      * congestion control
      *
      */
     void SetCACongestionSettingForAllNodes();

     /**
      * Prints to osm log the content of the given
      *        CC_CongestionControlTable struct
      *
      * @param cc_congestion_control_table struct to be printed
      */
     void DumpCCTableSetting(struct CC_CongestionControlTable& cc_congestion_control_table);

#if 0
     /**
      * Prints to osm log the content of the given
      *        CCTIEntryListVec struct
      *
      * @param ccti_list struct to be printed
      */
     void DumpCCTIList(const CCTIEntryListVec &ccti_list);
#endif

     /**
      * Calculates a CCTI list with 64 * control_table_cap entries
      *
      * @param control_table_cap number of 64 entries
      * @param ccti_list CCTIEntryListVec struct with the calculated
      *                  list
      *
      * @return int 0 if succeeded, otherwise 1
      */

     int GenerateCCTI(const unsigned int control_table_cap,
             CCTIEntryListVec& ccti_list);

     /**
      * Get default CCTI list. If list given by user returns
      *        that list, otherwise calculates a list with 64
      *         * control_table_cap entries based on mtu size
      *
      * @param control_table_cap control table cap of CA node
      * @param ccti_list cct list
      *
      * @return int 0 if succeeded, otherwise 1
      */
     int GetDefaultCANodeCCTIList(const unsigned int control_table_cap,
             CCTIEntryListVec& ccti_list);

     /**
      * Get settings for a specific CA node. Content will be
      *        filled in the given structs
      *
      * @param node_info node info
      * @param node_supports_cc true if node supports congestion
      *                         control, otherwise false
      * @param ccti_list CCTIEntryListVec with node settings (only if
      *                  succeeded)
      * @param node_it node iterator to database
      *
      * @return int 0 if succeeded, otherwise 1
      */
     int GetCANodeCCTIList(const CCNodeInfo& node_info,
             bool& node_supports_cc,
             CCTIEntryListVec& ccti_list,
             map<u_int64_t, CACCSettingDataBaseNodeEntry>::iterator& ca_node_it);

     /**
      * Compare CCTIEntryListVec
      * @param lhs
      * @param rhs
      *
      * @return bool true if all fields are equal, otherwise false
      */
     bool CompareCACongestionControlTableSetting(const CCTIEntryListVec &lhs,
             const CCTIEntryListVec &rhs);

     /**
      * Set CA CongestionControlTable
      * Sends CongestionControlTable MAD
      *
      * @param node_info node info
      *
      * @return int 0 if succeeded, otherwise 1
      */
     int SetCACongestionControlTable(const CCNodeInfo& node_info);

     /**
      * Set CA CongestionControlTable for all nodes that support
      * congestion control
      *
      */
     void SetCACongestionControlTableForAllNodes();

     /**
      * Prints to osm log the content of the given
      *        CC_SwitchCongestionSetting struct
      *
      * @param cc_sw_congestion_setting struct to be printed
      */
     void DumpSWCongSetting(struct CC_SwitchCongestionSetting& cc_sw_congestion_setting);

     /**
      * Set a bit in the mask with 1
      *
      * @param bit bit to set. range is [0,256)
      */
     void SetMask(UINT256& mask, u_int8_t bit);

     /**
      * Generate victim mask for Switch node. Set ports in victim
      * mask list that are connected to a CA
      *
      * @param node_info node info
      * @param victim_mask_list victim mask
      *
      * @return int 0 if succeeded, otherwise 1
      */
     int GenerateVictimMaskList(const CCNodeInfo& node_info,
             UINT256& victim_mask_list);

     /**
      * Get settings for a specific CA node. Content will be
      *        filled in the  given structs
      *
      * @param node_info node info
      * @param node_supports_cc true if node supports congestion
      *                         control, otherwise false
      * @param node_params CC_SwitchCongestionSetting struct with
      *                    node settings (only if succeeded)
      * @param port_params CC_SwitchPortCongestionSetting struct with
      *                    node settings (only if succeeded)
      *
      * @return int 0 if succeeded, otherwise 1
      */
     int GetSWNodeCCSettings(const CCNodeInfo& node_info,
             bool& node_supports_cc,
             struct CC_SwitchCongestionSetting& node_params,
             struct CC_SwitchPortCongestionSetting port_params[NUM_OF_32_PORTS],
             map<u_int64_t, SWCCSettingDataBaseNodeEntry>::iterator& sw_node_it);

     /**
      * Compare SwitchCongestionSetting
      *
      * @param lhs
      * @param rhs
      *
      * @return bool true if all fields are equal, otherwise false
      */
     bool CompareSwitchCongestionSetting(const struct CC_SwitchCongestionSetting &lhs,
             const struct CC_SwitchCongestionSetting &rhs);

     /**
      * Set Switch CongestionSetting
      * Sends SwitchCongestionSetting MAD
      *
      * @param node_info node info
      *
      * @return int 0 if succeeded, otherwise 1
      */
     int SetSwitchCongestionSetting(const CCNodeInfo& node_info);

     /**
      * Set Switch CongestionSetting for all nodes that support
      * congestion control
      *
      */
     void SetSwitchCongestionSettingForAllNodes();

     /**
      * Print to osm log the content of the given
      *        CC_SwitchPortCongestionSetting  struct
      *
      * @param block_idx index of the port block
      * @param cc_sw_port_congestion_setting struct to be printed
      */
     void DumpSWPortCongSetting(const unsigned int block_idx,
             struct CC_SwitchPortCongestionSetting& cc_sw_port_congestion_setting);

     /**
      * Set Switch PortCongestionSetting
      * Sends SwitchPortCongestionSetting MAD
      *
      * @param node_info node info
      *
      * @return int 0 if succeeded, otherwise 1
      */
     int SetSwitchPortCongestionSetting(const CCNodeInfo& node_info);

     /**
      * Set Switch PortCongestionSetting for all nodes that support
      * congestion control
      *
      */
     void SetSwitchPortCongestionSettingForAllNodes();

     /**
      * Set all nodes that suppport congestion control with given
      * congestion control parameters
      *
      *
      * @return int 0 if succeeded, otherwise 1
      */
     void SetCCParamsForAllNodes();

     ////////////////////
     // methods statistic agent
     ////////////////////
     /**
      * Create printable string of the given UINT256 struct
      *
      * @param mask_name struct name, used as a prefix to the struct
      *                  content
      * @param mask_st struct to be printed
      *
      * @return string
      */
     string DumpMask(const string &mask_name, const UINT256 &mask_st);

     /**
      * Create printable string of CA CC log
      *
      * @param cc_congestion_log_ca
      *
      * @return string
      */
     string DumpCACongestionLog(struct CC_CongestionLogCA& cc_congestion_log_ca);

     /**
      * Create printable string of Switch CC log
      *
      * @param cc_congestion_log_sw
      *
      * @return string
      */
     string DumpSWCongestionLog(struct CC_CongestionLogSwitch& cc_congestion_log_sw);

     /**
      * Get Switch CC Statistics.
      * CC Logs
      *
      * @param node_info node info
      *
      * @return int 0 if succeeded, otherwise 1
      */
     int GetSWCCStatistics(CCNodeInfo& node_info);

     /**
      * Get CA CC Statistics
      *
      * @param node_info node info
      *
      * @return int 0 if succeeded, otherwise 1
      */
     int GetCACCStatistics(CCNodeInfo& node_info);

     /**
      * Count number of congested ports on given switch.
      *
      * @param node_info node info. uses number of ports
      * @param cc_congestion_log_sw current congestion log from
      *                             switch
      *
      * @return unsigned int return number of ports congested in
      *         switch
      */
     unsigned int GetSWNumCongPorts(CCNodeInfo& node_info,
             struct CC_CongestionLogSwitch &cc_congestion_log_sw);

     /**
      * Collect CC logs from all CAs
      *
      * @return int 0 if succeeded, otherwise 1
      * @throw int If m_cc_mgr_request is true
      */
     int CollectCCStatisticsFromAllCAs();

     /**
      * Collect CC logs from all Switches
      *
      * @return int 0 if succeeded, otherwise 1
      * @throw int If m_cc_mgr_request is true
      */
     int CollectCCStatisticsFromAllSwitches();

     /**
      * After each cycle CC Statistics agent will sleep for certain period
      *
      * @param time_limit
      */
     void CCStatisticsSleep(time_t seconds);

     /**
      * Report Congestion Control Statistics
      *
      * @return int
      */
     int ReportCCStatistics();

     // This is the static class function that serves as a
     // C style function pointer for the pthread_create call
     /**
      * Poll periodically for CC logs from all nodes.
      *
      * return when m_run_cc_log_agent is false
      */
     void RunCCStatisticsAgent();

     /**
      * Create CC Statistics Agent
      *
      */
     void CreateCCStatisticsAgent();

     /**
      * Destroy CC Statistics Agent
      *
      */
     void DestroyCCStatisticsAgent();

     ////////////////////
     // methods FIT
     ////////////////////
     /**
      * Report messages to FabricIT
      *
      * @param msg_type message type
      * @param cnt number of elements in message data
      * @param msg_data pointer to message data
      *
      * @return int int 0 if succeeded, otherwise 1
      */
     int ReportFit(fit_msg_type_t msg_type, const int cnt, void* msg_data);

     ////////////////////
     // methods input file
     ////////////////////
     /**
      * Reset user option database. Clear old entries
      *
      */
     void ResetUserOptionsDataBase();

     /**
      * Split a comma separated list in to a vector of strings
      *
      * @param s a comma separated list
      *
      * @return vector<string> vector of strings
      */
     vector<string> Split(const string& s);

     /**
      * Check if given string is a numeric zero
      *
      * @param str_value string value
      *
      * @return bool Return true if string is a numerical zero,
      *         otherwise false
      */
     bool CheckStr(const string& str_value);

     /**
      * Check if this option was already set by user. Sets an
      *        indication in user_options_flags
      *
      * @param key user option key
      * @param user_options_flags map of user options that where set
      *
      * @return bool return true if key wasn't already set, otherwise
      *         false
      */
     bool CheckAndSetUserOptionFlag(const string& key,
             std::map<string, bool>& user_options_flags);

     /**
      * Check if enable user option was set to false and also other
      * flags were set
      *
      * @param enable
      * @param is_set_flags is set flags
      * @param key_str_type type of key. Switch, CA or General
      *
      * @return bool true if found conflict, otherwise false
      */
     bool CheckEnableConflict(bool enable,
             const map<string, bool> &is_set_flags,
             const string &key_str_type);

     /**
      * Set general user option to user option database
      *
      * @param key user option key
      * @param str_value string value
      *
      * @return int 0 if succeeded, otherwise 1
      */
     int SetGeneralUserOptions(const string &key, const string &str_value);

     /**
      * Set CA user option to user option database
      *
      * @param key user option key
      * @param str_value string value
      * @param cc_ca_congestion_setting CA CC_CACongestionSetting struct to set
      *                                 value in
      * @param ccti_list CA CCTIEntryListVec struct to set value in
      *                  case key is "cct"
      * @param is_set_flags map to set indication that user option
      *                     key was set
      * @param enable set value of "enable" user option key
      *
      * @return int 0 if succeeded, otherwise 1
      */
     int SetCAUserOption(const string &key, const string &str_value,
             struct CC_CACongestionSetting& cc_ca_congestion_setting,
             CCTIEntryListVec& ccti_list,
             map<string, bool>& is_set_flags,
             bool& enable);

     /**
      * Set Switch user option to user option database
      *
      * @param key user option key
      * @param str_value string value
      * @param cc_sw_congestion_setting Switch
      *                                 CC_SwitchCongestionSetting
      *                                 struct to set value in
      * @param is_set_flags map to set indication that user option
      *                     key was set
      * @param enable set value of "enable" user option key
      *
      * @return int 0 if succeeded, otherwise 1
      */
     int SetSWUserOption(const string &key, const string &str_value,
             struct CC_SwitchCongestionSetting& cc_sw_congestion_setting,
             map<string, bool>& is_set_flags,
             bool& enable);

     /**
      * Check if file exists and have permissions
      *
      * @return true if exists, otherwise false
      */
     bool IsFileExists(const char *file_name);

     /**
      * Parse user input file. Default location is
      *        /usr/etc/opensm/cc_mgr.conf or value of environment
      *        variable CC_MGR_CONF
      *
      * @return int 0 if succeeded, otherwise 1
      */
     int ParseInputFile();

     /**
      * Set general key value from user option to settings
      *        database
      *
      * @param key user option key
      */
     void SetGeneralOptionsFromUser(const string &key);

     /**
      * Set Switch user option from source settings to
      *        destination settings
      *
      * @param key user option key
      * @param src_params source CC_SwitchCongestionSetting struct
      * @param dest_params destination CC_SwitchCongestionSetting
      *                    struct
      * @param enable true if node supports congestion
      *                         control, otherwise false
      */
     void SetSWOptionsFromUser(const string &key,
             const struct CC_SwitchCongestionSetting src_params,
             struct CC_SwitchCongestionSetting& dest_params,
             bool& enable);

     /**
      * Set CA user option from source settings to destination
      *        settings
      *
      * @param key user option key
      * @param src_params source CC_CACongestionSetting struct
      * @param src_ccti_list source CCTIEntryListVec struct
      * @param dest_params dest CC_CACongestionSetting struct
      * @param dest_ccti_list dest CCTIEntryListVec struct
      * @param enable true if node supports congestion
      *                         control, otherwise false
      */
     void SetCAOptionsFromUser(const string &key,
             const struct CC_CACongestionSetting &src_params,
             const CCTIEntryListVec &src_ccti_list,
             struct CC_CACongestionSetting& dest_params,
             CCTIEntryListVec& dest_ccti_list,
             bool& enable);

     /**
      * Count number of hosts in subnet
      *
      * @return unsigned int return number of hosts in subnet
      */
     unsigned int GetNumHosts();

     /**
      * Quatization of number of hosts. Change number of hosts
      *        to the nearest power of 2 greater than or equal to
      *        given number of hosts. If number lower then
      *        CC_GENERAL_DEFAULT_MIN_HOSTS set num_hosts to this
      *        value
      *
      * @param num_hosts number of hosts
      */
     void NumHostsQuatization(unsigned int& num_hosts);

     /**
      * Set CC_SwitchCongestionSetting struct with parameters to
      * disable congestion control
      *
      * @param cc_sw_congestion_setting
      */
     void DisableSWParams(struct CC_SwitchCongestionSetting& cc_sw_congestion_setting);

     /**
      * Set CC_CACongestionSetting struct with parameters to disable
      * congestion control
      *
      * @param cc_ca_congestion_setting
      */
     void DisableCAParams(struct CC_CACongestionSetting& cc_ca_congestion_setting);

     /**
      * Set CACongestionEntryListElement struct with parameters to disable
      * congestion control
      *
      * @param cc_ca_cong_entry
      */
     void DisableCAEntryListElementParams(
        struct CACongestionEntryListElement& cc_ca_cong_entry);

     /**
      * Check if ca node parameters are set.
      * Compares previous settings only if already set, otherwise
      * return false and sets m_is_congestion_settings_set
      *
      * @param prev_params previous parameters
      * @param ca_node
      *
      * @return bool true if current parameters are already set,
      *         otherwise false
      */
     bool IsCANodeParamSet(const struct CACCSettingDataBaseNodeEntry &prev_params,
             struct CACCSettingDataBaseNodeEntry& ca_node);

     /**
      * Check if sw node parameters are set. Compares previous
      * settings only if already set, otherwise return false and sets
      * m_is_congestion_settings_set
      *
      * @param prev_params previous parameters
      * @param sw_node
      *
      * @return bool true if current parameters are already set,
      *         otherwise false
      */
     bool IsSWNodeParamSet(const struct CC_SwitchCongestionSetting &prev_params,
             struct SWCCSettingDataBaseNodeEntry& sw_node);

     /**
      * Reset error window mechanism. Remove old errors
      *
      */
     void ResetErrorWindow();

     /**
      * Merge between default settings and user option settings
      *
      */
     void UpdateUserOptions();

     ////////////////////
     // methods static
     ////////////////////
     /**
     * Static callback function for the statistics agent
     *
     * @param cc_mgr_obj CongestionControlManager object
     *
     * @return void* the object
     */
    static void* run_cc_statistics_agent(void *cc_mgr_obj)
    {
        //All we do here is call the thread worker function
        reinterpret_cast<CongestionControlManager *>(cc_mgr_obj)->RunCCStatisticsAgent();
        return cc_mgr_obj;
    }

    /**
     * Compare between CA CongestionLogEvent entrys
     *
     * @param m_node_params left hand side CongestionLogEvent
     * @param rhs right hand side CongestionLogEvent
     *
     * @return int 0 - m_node_params = rhs, positive - m_node_params > rhs, negative -
     *         m_node_params < rhs
     */
    static int cmp_ca_log_event(const void* m_node_params, const void* rhs)
    {
        return ((*(CongestionLogEventListCAElement*)m_node_params).Timestamp -
                (*(CongestionLogEventListCAElement*)rhs).Timestamp);
    }

    /**
     * Compare between Switch CongestionLogEvent entrys
     *
     * @param m_node_params left hand side CongestionLogEvent
     * @param rhs right hand side CongestionLogEvent
     *
     * @return int 0 - m_node_params = rhs, positive - m_node_params > rhs, negative -
     *         m_node_params < rhs
     */
    static int cmp_sw_log_event(const void* m_node_params, const void* rhs)
    {
        return ((*(CongestionEntryListSwitchElement*)m_node_params).Timestamp -
                (*(CongestionEntryListSwitchElement*)rhs).Timestamp);
    }
};


#endif  /* CC_MGR_H_ */
